@@ -35,6 +35,9 @@
 // Date: 15 May 2022
 //***************************************************************************/
 
+#include "ojph_arch.h"
+#if defined(OJPH_ARCH_I386) || defined(OJPH_ARCH_X86_64)
+
 #include <climits>
 #include <immintrin.h>
 #include "ojph_defs.h"
@@ -287,4 +290,6 @@ namespace ojph {
       }
     }
   }
-}
\ No newline at end of file
+}
+
+#endif
