@@ -35,11 +35,13 @@
 // Date: 28 August 2019
 //***************************************************************************/
 
+#include "ojph_arch.h"
+#if defined(OJPH_ARCH_I386) || defined(OJPH_ARCH_X86_64)
+
 #include <climits>
 #include <cstdio>
 
 #include "ojph_defs.h"
-#include "ojph_arch.h"
 #include "ojph_mem.h"
 #include "ojph_params.h"
 #include "../codestream/ojph_params_local.h"
@@ -1000,3 +1002,5 @@ namespace ojph {
 
   } // !local
 } // !ojph
+
+#endif
